from uuid import uuid4

from openai.types.responses import (
    ResponseFunctionCallArgumentsDeltaEvent,
    ResponseFunctionToolCall,
    ResponseOutputItemAddedEvent,
    ResponseReasoningSummaryTextDeltaEvent,
    ResponseStreamEvent,
    ResponseTextDeltaEvent,
)

from grasp_agents.typing.completion_chunk import (
    CompletionChunk,
    CompletionChunkDelta,
    CompletionChunkDeltaToolCall,
)
from grasp_agents.typing.message import Role

# Union of response chunk event variants produced by OpenAI's Responses API.
# Kept here to avoid circular imports between converters and chunk_converters.
ResponseApiChunk = (
    ResponseReasoningSummaryTextDeltaEvent
    | ResponseFunctionCallArgumentsDeltaEvent
    | ResponseOutputItemAddedEvent
    | ResponseTextDeltaEvent
)


def from_api_completion_chunk(
    api_chunk: ResponseApiChunk, name: str | None = None
) -> CompletionChunk:
    if isinstance(api_chunk, ResponseReasoningSummaryTextDeltaEvent):
        delta = CompletionChunkDelta(
            reasoning_content=api_chunk.delta, role=Role.ASSISTANT
        )
        return CompletionChunk(
            item_id=api_chunk.item_id,
            model=None,
            name=name,
            system_fingerprint=None,
            usage=None,
            delta=delta,
            finish_reason=None,
            logprobs=None,
        )
    if isinstance(api_chunk, ResponseFunctionCallArgumentsDeltaEvent):
        delta = CompletionChunkDelta(
            tool_calls=[
                CompletionChunkDeltaToolCall(
                    id=None,
                    index=api_chunk.output_index,
                    tool_arguments=api_chunk.delta,
                    tool_name=None,
                )
            ],
            role=Role.ASSISTANT,
        )
        return CompletionChunk(
            item_id=api_chunk.item_id,
            model=None,
            name=name,
            system_fingerprint=None,
            usage=None,
            delta=delta,
            finish_reason=None,
            logprobs=None,
        )
    if isinstance(api_chunk, ResponseOutputItemAddedEvent) and isinstance(
        api_chunk.item, ResponseFunctionToolCall
    ):
        item = api_chunk.item
        func_delta = CompletionChunkDelta(
            tool_calls=[
                CompletionChunkDeltaToolCall(
                    id=item.call_id,
                    index=api_chunk.output_index,
                    tool_arguments=item.arguments,
                    tool_name=item.name,
                )
            ],
            role=Role.ASSISTANT,
        )
        return CompletionChunk(
            item_id=item.id,
            model=None,
            name=name,
            system_fingerprint=None,
            usage=None,
            delta=func_delta,
            finish_reason=None,
            logprobs=None,
        )
    if isinstance(api_chunk, ResponseTextDeltaEvent):
        delta = CompletionChunkDelta(content=api_chunk.delta)

        return CompletionChunk(
            item_id=api_chunk.item_id,
            model=None,
            name=name,
            system_fingerprint=None,
            delta=delta,
            finish_reason=None,
            logprobs=api_chunk.logprobs,
            usage=None,
        )
    supported_types = (
        ResponseReasoningSummaryTextDeltaEvent,
        ResponseFunctionCallArgumentsDeltaEvent,
        ResponseOutputItemAddedEvent,
        ResponseTextDeltaEvent,
    )
    supported_type_names = ", ".join(t.__name__ for t in supported_types)
    raise TypeError(
        f"Unsupported chunk event type: {type(api_chunk)!r}. "
        f"Supported chunk event types are: {supported_type_names}. "
        "Ensure that you pass a chunk produced by the OpenAI Responses API "
        "streaming interface (for example, a text, reasoning summary, function "
        "call arguments delta, or function tool call output item event)."
    )


def to_completion_chunk(completion_chunk: CompletionChunk) -> ResponseStreamEvent:
    raise NotImplementedError



import fnmatch
import logging
import os
from collections.abc import AsyncGenerator, AsyncIterator, Iterable, Mapping
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, ClassVar

from openai import AsyncOpenAI
from openai._types import omit
from openai.lib.streaming.responses._responses import (
    AsyncResponseStreamManager,
)
from openai.types.responses import (
    ParsedResponse,
    Response,
    ResponseCompletedEvent,
    ResponseInputParam,
    ResponseOutputItemDoneEvent,
    ResponseStreamEvent,
    ResponseTextConfigParam,
)
from openai.types.responses.response_create_params import (
    StreamOptions as ResponsesStreamOptionsParam,
)
from openai.types.responses.response_create_params import (
    ToolChoice as ResponseToolChoice,
)
from openai.types.responses.tool_param import (
    ToolParam as ResponsesToolParam,
)
from openai.types.shared import Reasoning
from pydantic import BaseModel

from grasp_agents.cloud_llm import APIProvider, CloudLLM, CloudLLMSettings
from grasp_agents.typing.events import CompletionChunkEvent, CompletionItemEvent
from grasp_agents.typing.tool import BaseTool

from .converters import OpenAIResponsesConverters, ResponseApiChunk

logger = logging.getLogger(__name__)


class OpenAIResponsesLLMSettings(CloudLLMSettings, total=False):
    # web search should be put as a tool:
    # tools=[{
    #   "type": "web_search_preview",
    #   "search_context_size": "high",  # Options: "low", "medium", "high"
    #    "user_location": {...}
    # }]
    reasoning: Reasoning
    parallel_tool_calls: bool
    max_output_tokens: int
    top_logprobs: int | None

    text: ResponseTextConfigParam
    stream_options: ResponsesStreamOptionsParam | None
    store: bool | None
    user: str


@dataclass(frozen=True)
class OpenAIResponsesLLM(CloudLLM):
    llm_settings: OpenAIResponsesLLMSettings | None = None
    converters: ClassVar[OpenAIResponsesConverters] = OpenAIResponsesConverters()
    openai_client_timeout: float = 120.0
    openai_client_max_retries: int = 2
    extra_openai_client_params: dict[str, Any] | None = None
    client: AsyncOpenAI = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        _model_name = self.model_name
        _openai_client_params = deepcopy(self.extra_openai_client_params or {})
        _openai_client_params["timeout"] = self.openai_client_timeout
        _openai_client_params["max_retries"] = self.openai_client_max_retries
        if self.http_client is not None:
            _openai_client_params["http_client"] = self.http_client

        _client = AsyncOpenAI(
            base_url="https://api.openai.com/v1",
            api_key=os.getenv("OPENAI_API_KEY"),
            **_openai_client_params,
        )

        object.__setattr__(self, "model_name", _model_name)
        object.__setattr__(self, "client", _client)

    async def _get_api_completion(
        self,
        api_messages: list[ResponseInputParam],
        *,
        api_tools: list[ResponsesToolParam] | None = None,
        api_tool_choice: ResponseToolChoice | None = None,
        api_response_schema: type[Any] | None = None,
        **api_llm_settings: Any,
    ) -> ParsedResponse[Any] | Response:
        messages = [subitem for item in api_messages for subitem in item]
        tools = api_tools or []
        tool_choice = api_tool_choice if api_tool_choice is not None else omit
        text_format = api_response_schema if api_response_schema is not None else omit
        response_id = api_llm_settings.get("previous_response_id")
        if self.apply_response_schema_via_provider:
            return await self.client.responses.parse(
                model=self.model_name,
                input=[messages[-1]] if response_id else messages,
                tools=tools,
                tool_choice=tool_choice,
                text_format=text_format,
                **api_llm_settings,
            )
        return await self.client.responses.create(
            model=self.model_name,
            input=[messages[-1]] if response_id else messages,
            tools=tools,
            tool_choice=tool_choice,
            stream=False,
            **api_llm_settings,
        )

    async def _get_api_completion_stream(
        self,
        api_messages: list[ResponseInputParam],
        api_tools: Iterable[ResponsesToolParam] | None = None,
        api_tool_choice: ResponseToolChoice | None = None,
        api_response_schema: type[Any] | None = None,
        **api_llm_settings: Any,
    ) -> AsyncIterator[ResponseStreamEvent]:
        messages = [subitem for item in api_messages for subitem in item]
        tools = api_tools if api_tools is not None else omit
        response_id = api_llm_settings.get("previous_response_id")
        tool_choice = api_tool_choice if api_tool_choice is not None else omit
        text_format = api_response_schema if api_response_schema is not None else omit
        _api_llm_settings = dict(api_llm_settings)
        if "stream_options" in _api_llm_settings:
            so = dict(_api_llm_settings.get("stream_options") or {})
            so.pop("include_usage", None)
            _api_llm_settings["stream_options"] = so

        async def iterator() -> AsyncIterator[ResponseStreamEvent]:
            effective_text_format = (
                text_format if self.apply_response_schema_via_provider else omit
            )
            stream_manager: AsyncResponseStreamManager[Any] = (
                self.client.responses.stream(
                    model=self.model_name,
                    input=[messages[-1]] if response_id else messages,
                    tool_choice=tool_choice,
                    tools=tools,
                    text_format=effective_text_format,
                    **_api_llm_settings,
                )
            )

            async with stream_manager as stream:
                async for response_event in stream:
                    yield response_event

        return iterator()

    def combine_completion_chunks(
        self,
        completion_chunks: list[ResponseStreamEvent],
        response_schema: Any | None = None,
        tools: Mapping[str, BaseTool[BaseModel, Any, Any]] | None = None,
    ) -> Response:
        final_resp = None
        if len(completion_chunks) > 0:
            final_resp = completion_chunks[-1]
            if isinstance(final_resp, ResponseCompletedEvent):
                return final_resp.response
        raise RuntimeError("No 'response.completed' event received")

    async def _handle_api_stream_event(  # type: ignore
        self,
        event: ResponseStreamEvent,
        proc_name: str | None = None,
        call_id: str | None = None,
    ) -> AsyncGenerator[CompletionChunkEvent[Any] | CompletionItemEvent, None]:
        if isinstance(event, ResponseApiChunk):
            try:
                completion_chunk = self.converters.from_completion_chunk(
                    event, name=self.model_id
                )
            except TypeError:
                logger.exception(
                    "Skipping chunk conversion for event type %s",
                    type(event).__name__,
                )
                return
            yield CompletionChunkEvent(
                data=completion_chunk, src_name=proc_name, call_id=call_id
            )
        if isinstance(event, ResponseOutputItemDoneEvent):
            try:
                completion_item = self.converters.from_api_item(
                    event, name=self.model_id
                )
            except TypeError:
                logger.exception(
                    "Failed to convert ResponseOutputItemDoneEvent "
                    "to CompletionItem (item type: %s)",
                    getattr(event.item, "type", "unknown"),
                )
                return
            yield CompletionItemEvent(
                data=completion_item, src_name=proc_name, call_id=call_id
            )



from litellm.types.llms.openai import (
    ChatCompletionAnnotation,
    ChatCompletionAnnotationURLCitation,
)
from openai.types.responses import (
    EasyInputMessageParam,
    ResponseFunctionToolCall,
    ResponseFunctionToolCallParam,
    ResponseInputParam,
    ResponseOutputMessage,
    ResponseOutputText,
    ResponseReasoningItem,
    ResponseReasoningItemParam,
)
from openai.types.responses import Response as OpenAIResponse
from openai.types.responses.response_input_item_param import FunctionCallOutput
from openai.types.responses.response_input_message_item import (
    ResponseInputMessageItem as OpenAIResponseInputMessage,
)
from openai.types.responses.response_reasoning_item_param import Summary

from grasp_agents.typing.content import Content, ContentPartText
from grasp_agents.typing.message import (
    AssistantMessage,
    RedactedThinkingBlock,
    SystemMessage,
    ThinkingBlock,
    ToolMessage,
    UserMessage,
)
from grasp_agents.typing.tool import ToolCall

from .content_converters import from_api_content, to_api_content


def from_api_responses_output_message(
    output: ResponseOutputMessage,
) -> tuple[list[str], list[ChatCompletionAnnotation], str | None]:
    content: list[str] = []
    refusal: str | None = None
    annotations: list[ChatCompletionAnnotation] = []
    raw_contents = output.content
    for raw_content in raw_contents:
        if isinstance(raw_content, ResponseOutputText):
            content.append(raw_content.text)
            if raw_content.annotations:
                annotations.extend(
                    [
                        ChatCompletionAnnotation(
                            type="url_citation",
                            url_citation=ChatCompletionAnnotationURLCitation(
                                **api_annotation.model_dump()
                            ),
                        )
                        for api_annotation in raw_content.annotations
                    ]
                )
        else:
            refusal = raw_content.refusal
    return content, annotations, refusal


def from_api_user_message(
    api_message: OpenAIResponseInputMessage, name: str | None = None
) -> UserMessage:
    content = from_api_content(api_message.content)
    return UserMessage(content=content, name=name)


def to_api_user_message(message: UserMessage) -> ResponseInputParam:
    input_message: ResponseInputParam = []
    api_content = (
        to_api_content(message.content)
        if isinstance(message.content, Content)
        else message.content
    )
    usr_message = EasyInputMessageParam(
        role="user", content=api_content, type="message"
    )
    input_message.append(usr_message)
    return input_message


def from_api_assistant_message(
    raw_completion: OpenAIResponse, name: str | None = None
) -> AssistantMessage:
    output_items = raw_completion.output
    content: list[str] = []
    refusal: str | None = None
    reasoning_id: str | None = None
    encrypted_content = None
    reasoning_summary: list[str] = []
    tool_calls: list[ToolCall] = []
    annotations: list[ChatCompletionAnnotation] = []
    for output in output_items:
        if isinstance(output, ResponseOutputMessage):
            content_output, annotation_output, refusal = (
                from_api_responses_output_message(output)
            )
            content.extend(content_output)
            annotations.extend(annotation_output)
        if isinstance(output, ResponseReasoningItem):
            raw_summaries = output.summary
            encrypted_content = output.encrypted_content
            reasoning_id = output.id
            for raw_summary in raw_summaries:
                reasoning_summary.append(raw_summary.text)
        if isinstance(output, ResponseFunctionToolCall):
            tool = ToolCall(
                id=output.call_id,
                tool_arguments=output.arguments,
                tool_name=output.name,
            )
            tool_calls.append(tool)
    thinking_blocks: list[ThinkingBlock] = [
        ThinkingBlock(type="thinking", thinking=item) for item in reasoning_summary
    ]
    if encrypted_content and thinking_blocks:
        thinking_blocks[0]["signature"] = encrypted_content
    return AssistantMessage(
        content=" ".join(content),
        annotations=annotations,
        reasoning_content=" ".join(reasoning_summary),
        tool_calls=tool_calls,
        refusal=refusal,
        response_id=raw_completion.id,
        thinking_blocks=thinking_blocks,
        reasoning_id=reasoning_id,
    )


def to_api_assistant_message(
    message: AssistantMessage,
) -> ResponseInputParam:
    input_message: ResponseInputParam = []
    if message.content:
        text_message = EasyInputMessageParam(
            role="assistant", content=message.content, type="message"
        )
        input_message.append(text_message)
    if message.tool_calls:
        tool_calls = [
            ResponseFunctionToolCallParam(
                arguments=item.tool_arguments,
                call_id=item.id,
                name=item.tool_name,
                type="function_call",
            )
            for item in message.tool_calls
        ]
        input_message.extend(tool_calls)
    if message.reasoning_id:
        reasoning = ResponseReasoningItemParam(
            id=message.reasoning_id,
            summary=[
                Summary(text=text, type="summary_text")
                for text in message.thinking_summaries
            ],
            type="reasoning",
            encrypted_content=message.encrypted_content,
        )
        input_message.append(reasoning)
    return input_message


def from_api_system_message(
    api_message: OpenAIResponseInputMessage, name: str | None = None
) -> SystemMessage:
    content_obj = from_api_content(api_message.content)
    content_str = "".join(
        part.data for part in content_obj.parts if isinstance(part, ContentPartText)
    )
    return SystemMessage(content=content_str, name=name)


def to_api_system_message(message: SystemMessage) -> ResponseInputParam:
    input_message: ResponseInputParam = []
    sys_message = EasyInputMessageParam(
        role="system", content=message.content or "", type="message"
    )
    input_message.append(sys_message)
    return input_message


def from_api_tool_message(
    api_message: FunctionCallOutput, name: str | None = None
) -> ToolMessage:
    return ToolMessage(
        content=api_message["output"],  # type: ignore
        tool_call_id=api_message["call_id"],
        name=name,
    )


# this is conversion from call output result
def to_api_tool_message(message: ToolMessage) -> ResponseInputParam:
    input_message: ResponseInputParam = []
    input_message.append(
        FunctionCallOutput(
            type="function_call_output",
            output=message.content,
            call_id=message.tool_call_id,
        )
    )
    return input_message



from openai.types.shared import Reasoning as OpenAIReasoning
from openai.types.responses.response_create_params import (
    StreamOptions as OpenAIResponsesStreamOptionsParam,
)

from .completions import OpenAILLM, OpenAILLMSettings
from .responses import OpenAIResponsesLLM, OpenAIResponsesLLMSettings

__all__ = [
    "OpenAILLM",
    "OpenAILLMSettings",
    "OpenAIResponsesLLM",
    "OpenAIResponsesLLMSettings",
    "OpenAIReasoning",
    "OpenAIResponsesStreamOptionsParam",
]



from __future__ import annotations

from typing import Any

from openai.types.responses import Response as OpenAIResponse
from openai.types.responses import (
    ResponseInputParam,
    ResponseOutputItemDoneEvent,
    ResponseUsage,
)
from openai.types.responses.response_create_params import (
    ToolChoice as OpenAIResponseToolChoice,
)
from openai.types.responses.response_input_item_param import FunctionCallOutput
from openai.types.responses.response_input_message_item import (
    ResponseInputMessageItem as OpenAIResponseInputMessage,
)
from openai.types.responses.tool_param import ToolParam as OpenAIResponseToolParam
from pydantic import BaseModel

from grasp_agents.typing.completion import Completion, Usage
from grasp_agents.typing.completion_chunk import CompletionChunk
from grasp_agents.typing.completion_item import CompletionItem
from grasp_agents.typing.content import Content
from grasp_agents.typing.converters import Converters
from grasp_agents.typing.message import (
    AssistantMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)
from grasp_agents.typing.tool import BaseTool, ToolChoice

from .chunk_converters import (
    ResponseApiChunk,
    from_api_completion_chunk,
    to_completion_chunk,
)
from .completion_converters import (
    completion_from_response,
    from_response_usage,
    to_api_completion,
)
from .completion_item_converters import from_api_completion_item
from .content_converters import from_api_content, to_api_content
from .message_converters import (
    from_api_assistant_message,
    from_api_system_message,
    from_api_tool_message,
    from_api_user_message,
    to_api_assistant_message,
    to_api_system_message,
    to_api_tool_message,
    to_api_user_message,
)
from .tool_converters import to_api_tool, to_api_tool_choice


class OpenAIResponsesConverters(Converters):
    @staticmethod
    def to_system_message(
        system_message: SystemMessage, **kwargs: Any
    ) -> ResponseInputParam:
        return to_api_system_message(system_message, **kwargs)

    @staticmethod
    def to_user_message(user_message: UserMessage, **kwargs: Any) -> ResponseInputParam:
        return to_api_user_message(user_message, **kwargs)

    @staticmethod
    def to_assistant_message(
        assistant_message: AssistantMessage, **kwargs: Any
    ) -> ResponseInputParam:
        return to_api_assistant_message(assistant_message, **kwargs)

    @staticmethod
    def from_system_message(
        raw_message: OpenAIResponseInputMessage,
        **kwargs: Any,
    ) -> SystemMessage:
        return from_api_system_message(raw_message, **kwargs)

    @staticmethod
    def from_user_message(
        raw_message: OpenAIResponseInputMessage,
        **kwargs: Any,
    ) -> UserMessage:
        return from_api_user_message(raw_message, **kwargs)

    @staticmethod
    def from_assistant_message(
        raw_message: OpenAIResponse,
        **kwargs: Any,
    ) -> AssistantMessage:
        return from_api_assistant_message(raw_message, **kwargs)

    @staticmethod
    def to_tool(
        tool: BaseTool[BaseModel, Any, Any], strict: bool | None = None, **kwargs: Any
    ) -> OpenAIResponseToolParam:
        return to_api_tool(tool=tool, strict=strict, **kwargs)

    @staticmethod
    def to_tool_choice(
        tool_choice: ToolChoice, **kwargs: Any
    ) -> OpenAIResponseToolChoice:
        return to_api_tool_choice(tool_choice=tool_choice, **kwargs)

    @staticmethod
    def to_tool_message(tool_message: ToolMessage, **kwargs: Any) -> ResponseInputParam:
        return to_api_tool_message(tool_message, **kwargs)

    @staticmethod
    def from_tool_message(
        raw_message: FunctionCallOutput, **kwargs: Any
    ) -> ToolMessage:
        return from_api_tool_message(raw_message, **kwargs)

    @staticmethod
    def to_content(content: Content, **kwargs: Any) -> Any:
        return to_api_content(content, **kwargs)

    @staticmethod
    def from_content(raw_content: Any, **kwargs: Any) -> Content:
        return from_api_content(raw_content, **kwargs)

    @staticmethod
    def to_completion(completion: Completion, **kwargs: Any) -> Any:
        return to_api_completion(completion, **kwargs)

    @staticmethod
    def from_completion(
        raw_completion: OpenAIResponse,
        name: str | None = None,
        **kwargs: Any,
    ) -> Completion:
        return completion_from_response(raw_completion, name=name, **kwargs)

    @staticmethod
    def from_completion_usage(raw_usage: ResponseUsage, **kwargs: Any) -> Usage:
        return from_response_usage(raw_usage, **kwargs)

    @staticmethod
    def to_completion_chunk(chunk: CompletionChunk, **kwargs: Any) -> Any:
        return to_completion_chunk(chunk, **kwargs)

    @staticmethod
    def from_completion_chunk(
        raw_chunk: ResponseApiChunk, name: str | None = None, **kwargs: Any
    ) -> CompletionChunk:
        return from_api_completion_chunk(raw_chunk, name=name, **kwargs)

    @staticmethod
    def from_api_item(
        raw_event: ResponseOutputItemDoneEvent, name: str | None = None, **kwargs: Any
    ) -> CompletionItem:
        return from_api_completion_item(raw_event, name=name, **kwargs)



from collections.abc import Iterable

from openai.types.responses import (
    ResponseInputContent,
    ResponseInputImage,
    ResponseInputImageParam,
    ResponseInputMessageContentListParam,
    ResponseInputText,
    ResponseInputTextParam,
)

from ...typing.content import (
    Content,
    ContentPart,
    ContentPartImage,
    ContentPartText,
    ImageData,
)
from ..completions.content_converters import BASE64_PREFIX, image_data_to_str


def from_api_content(
    api_content: str | Iterable[ResponseInputContent],
) -> "Content":
    if isinstance(api_content, str):
        return Content(parts=[ContentPartText(data=api_content)])

    content_parts: list[ContentPart] = []
    for api_content_part in api_content:
        content_part: ContentPart

        if isinstance(api_content_part, ResponseInputText):
            text_data = api_content_part.text
            content_part = ContentPartText(data=text_data)

        elif isinstance(api_content_part, ResponseInputImage):
            if (
                api_content_part.image_url is None
                or not api_content_part.image_url.startswith(BASE64_PREFIX)
            ):
                raise ValueError(f"Unsupported image url: {api_content_part.image_url}")
            url = api_content_part.image_url or ""
            detail = api_content_part.detail
            if url.startswith(BASE64_PREFIX):
                image_data = ImageData.from_base64(
                    base64_encoding=url.removeprefix(BASE64_PREFIX),
                    detail=detail,
                )
            else:
                image_data = ImageData.from_url(img_url=url, detail=detail)
            content_part = ContentPartImage(data=image_data)

        content_parts.append(content_part)  # type: ignore

    return Content(parts=content_parts)


def to_api_content(content: Content) -> ResponseInputMessageContentListParam:
    api_content: ResponseInputMessageContentListParam = []
    for content_part in content.parts:
        if isinstance(content_part, ContentPartText):
            api_content_part = ResponseInputTextParam(
                type="input_text", text=content_part.data
            )
        else:
            api_content_part = ResponseInputImageParam(
                type="input_image",
                detail=content_part.data.detail,
                image_url=image_data_to_str(content_part.data),
            )
        api_content.append(api_content_part)

    return api_content



from typing import Any

from openai import pydantic_function_tool
from openai.lib._pydantic import to_strict_json_schema
from openai.types.chat.chat_completion_tool_param import (
    ChatCompletionToolParam as OpenAIChatToolParam,
)
from openai.types.responses.response_create_params import (
    ToolChoice as OpenAIResponseToolChoice,
)
from openai.types.responses.tool_param import ToolParam as OpenAIResponseToolParam
from pydantic import BaseModel

from ...typing.tool import BaseTool, NamedToolChoice, ToolChoice


def to_api_tool(
    tool: BaseTool[BaseModel, Any, Any], strict: bool | None = None
) -> OpenAIResponseToolParam:
    return {
        "type": "function",
        "name": tool.name,
        "description": tool.description,
        "parameters": to_strict_json_schema(tool.in_type),
        "strict": True if strict is None else strict,
    }


def to_api_tool_choice(tool_choice: ToolChoice) -> OpenAIResponseToolChoice:
    if isinstance(tool_choice, NamedToolChoice):
        return {"type": "function", "name": tool_choice.name}
    return tool_choice  # type: ignore[return-value]



from __future__ import annotations

from openai.types.responses import Response as OpenAIResponse
from openai.types.responses import ResponseUsage

from grasp_agents.typing.completion import Completion, Usage

from .message_converters import from_api_assistant_message


def from_response_usage(raw_usage: ResponseUsage) -> Usage:
    return Usage(
        input_tokens=raw_usage.input_tokens,
        output_tokens=raw_usage.output_tokens
        - raw_usage.output_tokens_details.reasoning_tokens,
        reasoning_tokens=raw_usage.output_tokens_details.reasoning_tokens,
        cached_tokens=raw_usage.input_tokens_details.cached_tokens,
    )


def completion_from_response(
    raw_completion: OpenAIResponse,
    *,
    name: str | None = None,
) -> Completion:
    message = from_api_assistant_message(raw_completion)
    return Completion(
        id=raw_completion.id,
        model=raw_completion.model,
        created=int(raw_completion.created_at),
        message=message,
        usage=from_response_usage(raw_completion.usage)
        if raw_completion.usage
        else None,
        finish_reason=None,
        name=name,
    )


def to_api_completion(completion: Completion) -> OpenAIResponse:
    raise NotImplementedError



import contextlib
import inspect
import json
import os
import pathlib
import traceback
import types
import warnings
from collections.abc import Callable
from functools import wraps
from logging import getLogger
from typing import Any, TypeVar, cast

from opentelemetry import context as context_api
from opentelemetry import trace
from opentelemetry.instrumentation.utils import suppress_instrumentation
from opentelemetry.semconv_ai import SpanAttributes, TraceloopSpanKindValues
from opentelemetry.trace.status import Status, StatusCode
from pydantic import BaseModel
from traceloop.sdk.tracing.tracing import (
    TracerWrapper,
    get_chained_entity_path,
    set_entity_path,
    set_workflow_name,
)
from traceloop.sdk.utils import camel_to_snake  # type: ignore[import]
from traceloop.sdk.utils.json_encoder import JSONEncoder

logger = getLogger(__name__)

DEFAULT_EXCLUDE_FIELDS = {"_hidden_params", "completions"}
_CTX_DISABLE_TRACING_KEY = "override_disable_tracing"

T = TypeVar("T", bound=type)
F = TypeVar("F", bound=Callable[..., Any])


def _to_plain(obj: Any, exclude_fields: set[str] | None = None) -> Any:
    """
    Recursively convert objects to JSON-serializable primitives.

    - Pydantic BaseModel -> dict via model_dump()
    - dict/list/tuple/set -> recurse (sets become lists)
    - other objects -> returned as-is (left to JSONEncoder)
    """
    all_exclude_fields = DEFAULT_EXCLUDE_FIELDS.union(exclude_fields or set())
    if isinstance(obj, BaseModel):
        try:
            return obj.model_dump(exclude=all_exclude_fields)
        except Exception:
            return str(obj)

    if isinstance(obj, dict):
        items_dict = cast("dict[Any, Any]", obj)
        result: dict[str, Any] = {}
        for k, v in items_dict.items():
            if str(k) not in all_exclude_fields:
                result[str(k)] = _to_plain(v, exclude_fields)
        return result

    if isinstance(obj, (tuple, list, set)):
        items = cast("tuple[Any, ...] | list[Any] | set[Any]", obj)
        lst: list[Any] = []
        for v in items:
            lst.append(_to_plain(v, exclude_fields))
        return lst

    return obj


def _truncate_json_if_needed(json_str: str) -> str:
    """
    Truncate JSON string if it exceeds OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT;
    truncation may yield an invalid JSON string, which is expected for logging purposes.
    """
    limit_str = os.getenv("OTEL_SPAN_ATTRIBUTE_VALUE_LENGTH_LIMIT")

    if limit_str:
        try:
            limit = int(limit_str)
            if limit > 0 and len(json_str) > limit:
                return json_str[:limit]
        except ValueError:
            pass

    return json_str


def _should_send_prompts():
    return (
        os.getenv("TRACELOOP_TRACE_CONTENT") or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


def _get_span_name(
    entity_name: str,
    tlp_span_kind: TraceloopSpanKindValues,
    instance: type | None = None,
    kwargs: dict[str, Any] | None = None,
):
    if tlp_span_kind in {
        TraceloopSpanKindValues.WORKFLOW,
        TraceloopSpanKindValues.AGENT,
    }:
        set_workflow_name(entity_name)

    instance_name = None

    if instance is not None:
        if tlp_span_kind in {
            TraceloopSpanKindValues.WORKFLOW,
            TraceloopSpanKindValues.AGENT,
            TraceloopSpanKindValues.TOOL,
        }:
            instance_name = getattr(instance, "name", None)

        elif (
            tlp_span_kind == TraceloopSpanKindValues.TASK and entity_name == "generate"
        ):
            instance_name = getattr(instance, "agent_name", None)

    if instance_name:
        call_id = (kwargs or {}).get("call_id")
        span_name = f"{instance_name}.{entity_name}" + (
            f"[{call_id}]" if call_id else ""
        )
    else:
        span_name = f"{entity_name}.{tlp_span_kind.value}"

    return span_name


def _set_span_attributes(
    span: trace.Span,
    entity_name: str,
    tlp_span_kind: TraceloopSpanKindValues,
    version: int | None,
):
    if tlp_span_kind in {
        TraceloopSpanKindValues.TASK,
        TraceloopSpanKindValues.TOOL,
    }:
        entity_path = get_chained_entity_path(entity_name)
        set_entity_path(entity_path)

    span.set_attribute(SpanAttributes.TRACELOOP_SPAN_KIND, tlp_span_kind.value)
    span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_NAME, entity_name)
    if version:
        span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_VERSION, version)


def _handle_span_input(
    span: trace.Span,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    cls: type[JSONEncoder] | None = None,
    exclude_fields: set[str] | None = None,
):
    try:
        if _should_send_prompts():
            json_input = json.dumps(
                {
                    "args": _to_plain(list(args), exclude_fields=exclude_fields),
                    "kwargs": _to_plain(kwargs, exclude_fields=exclude_fields),
                },
                cls=cls,
                indent=2,
            )
            truncated_json = _truncate_json_if_needed(json_input)
            span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_INPUT, truncated_json)

    except TypeError as e:
        span.record_exception(e)
        span.set_status(StatusCode.ERROR, str(e))


def _handle_span_output(
    span: trace.Span,
    res: Any,
    cls: type[JSONEncoder] | None = None,
    exclude_fields: set[str] | None = None,
):
    """Handles entity output logging in JSON for both sync and async functions"""
    try:
        if _should_send_prompts():
            json_output = json.dumps(
                _to_plain(res, exclude_fields=exclude_fields), cls=cls, indent=2
            )
            truncated_json = _truncate_json_if_needed(json_output)
            span.set_attribute(SpanAttributes.TRACELOOP_ENTITY_OUTPUT, truncated_json)

    except TypeError as e:
        span.record_exception(e)
        span.set_status(StatusCode.ERROR, str(e))


def is_bound_method(func: Callable[..., Any], self_candidate: Any) -> bool:
    return (inspect.ismethod(func) and (func.__self__ is self_candidate)) or hasattr(
        self_candidate, func.__name__
    )


def _is_async_method(fn: Callable[..., Any]) -> bool:
    return inspect.iscoroutinefunction(fn) or inspect.isasyncgenfunction(fn)


# Quiet wrapper that suppresses prints and warnings from TracerWrapper.verify_initialized
def _tracing_initialized_quietly() -> bool:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            with (
                pathlib.Path(os.devnull).open("w") as devnull,
                contextlib.redirect_stdout(devnull),
                contextlib.redirect_stderr(devnull),
            ):
                return TracerWrapper.verify_initialized()
        except Exception:
            return False


def _is_tracing_globally_disabled() -> bool:
    try:
        return bool(context_api.get_value(_CTX_DISABLE_TRACING_KEY))
    except Exception:
        return False


@contextlib.contextmanager
def _suppress_tracing_globally():
    token = context_api.attach(context_api.set_value(_CTX_DISABLE_TRACING_KEY, True))
    try:
        with suppress_instrumentation():
            yield
    finally:
        context_api.detach(token)


def _tracing_enabled(
    instance: type | None = None,
) -> bool:
    if _is_tracing_globally_disabled():
        return False
    if instance is None:
        return True
    flag = getattr(instance, "tracing_enabled", True)
    return bool(flag)


def _exclude_fields_from_instance(
    instance: type | None = None,
) -> set[str] | None:
    if instance is None:
        return None
    exclude_fields = getattr(instance, "tracing_exclude_input_fields", None)
    if exclude_fields is not None and isinstance(exclude_fields, (list, set, tuple)):
        return set(exclude_fields)
    return None


def entity_method(
    name: str | None = None,
    version: int | None = None,
    tlp_span_kind: TraceloopSpanKindValues = TraceloopSpanKindValues.TASK,
) -> Callable[[F], F]:
    def decorate(fn: F) -> F:
        is_async = _is_async_method(fn)
        entity_name = name or fn.__qualname__

        if is_async:
            if inspect.isasyncgenfunction(fn):

                @wraps(fn)
                async def async_gen_wrap(*args: Any, **kwargs: Any) -> Any:
                    is_bound = is_bound_method(fn, args[0] if args else False)
                    instance = args[0] if is_bound else None
                    input_args = args[1:] if is_bound else args

                    exclude_fields = _exclude_fields_from_instance(instance)
                    is_enabled = _tracing_enabled(instance)

                    if not (is_enabled and _tracing_initialized_quietly()):
                        with _suppress_tracing_globally():
                            async for item in fn(*args, **kwargs):
                                yield item
                            return

                    span_name = _get_span_name(
                        entity_name, tlp_span_kind, instance=instance, kwargs=kwargs
                    )
                    tracer = trace.get_tracer(__name__)

                    with tracer.start_as_current_span(span_name) as span:
                        _set_span_attributes(span, entity_name, tlp_span_kind, version)
                        _handle_span_input(
                            span,
                            input_args,
                            kwargs,
                            cls=JSONEncoder,
                            exclude_fields=exclude_fields,
                        )
                        items: list[Any] = []

                        try:
                            async for item in fn(*args, **kwargs):
                                items.append(item)
                                yield item

                        except Exception as e:
                            span.set_status(Status(StatusCode.ERROR, str(e)))
                            span.record_exception(
                                e,
                                attributes={"tb": traceback.format_exc()},
                            )
                            raise

                        finally:
                            if items:
                                _handle_span_output(span, items[-1], cls=JSONEncoder)

                return cast("F", async_gen_wrap)

            @wraps(fn)
            async def async_wrap(*args: Any, **kwargs: Any) -> Any:
                is_bound = is_bound_method(fn, args[0] if args else False)
                instance = args[0] if is_bound else None
                input_args = args[1:] if is_bound else args

                exclude_fields = _exclude_fields_from_instance(instance)
                is_enabled = _tracing_enabled(instance)

                if not (is_enabled and _tracing_initialized_quietly()):
                    with _suppress_tracing_globally():
                        return await fn(*args, **kwargs)
                span_name = _get_span_name(
                    entity_name,
                    tlp_span_kind,
                    instance=instance,
                    kwargs=kwargs,
                )

                tracer = trace.get_tracer(__name__)

                with tracer.start_as_current_span(span_name) as span:
                    _set_span_attributes(span, entity_name, tlp_span_kind, version)
                    _handle_span_input(
                        span,
                        input_args,
                        kwargs,
                        cls=JSONEncoder,
                        exclude_fields=exclude_fields,
                    )

                    try:
                        res = await fn(*args, **kwargs)
                        _handle_span_output(span, res, cls=JSONEncoder)
                        return res

                    except Exception as e:
                        span.set_status(Status(StatusCode.ERROR, str(e)))
                        span.record_exception(
                            e, attributes={"tb": traceback.format_exc()}
                        )
                        raise

            return cast("F", async_wrap)

        @wraps(fn)
        def sync_wrap(*args: Any, **kwargs: Any) -> Any:
            is_bound = is_bound_method(fn, args[0] if args else False)
            instance = args[0] if is_bound else None

            exclude_fields = _exclude_fields_from_instance(instance)
            is_enabled = _tracing_enabled(instance)
            input_args = args[1:] if is_bound else args

            if not (is_enabled and _tracing_initialized_quietly()):
                with _suppress_tracing_globally():
                    return fn(*args, **kwargs)
            span_name = _get_span_name(
                entity_name,
                tlp_span_kind,
                instance=instance,
                kwargs=kwargs,
            )

            tracer = trace.get_tracer(__name__)

            with tracer.start_as_current_span(span_name) as span:
                _set_span_attributes(span, entity_name, tlp_span_kind, version)
                _handle_span_input(
                    span,
                    input_args,
                    kwargs,
                    cls=JSONEncoder,
                    exclude_fields=exclude_fields,
                )
                items: list[Any] = []

                try:
                    res = fn(*args, **kwargs)
                    if isinstance(res, types.GeneratorType):
                        for item in res:  # type: ignore[union-attr]
                            items.append(item)
                            yield item
                    else:
                        _handle_span_output(span, res, cls=JSONEncoder)
                        return res

                except Exception as e:
                    span.set_status(Status(StatusCode.ERROR, str(e)))
                    span.record_exception(e, attributes={"tb": traceback.format_exc()})
                    raise

                finally:
                    if items:
                        _handle_span_output(span, items[-1], cls=JSONEncoder)

        return cast("F", sync_wrap)

    return decorate


def entity_class(
    name: str | None,
    version: int | None,
    method_name: str,
    tlp_span_kind: TraceloopSpanKindValues = TraceloopSpanKindValues.TASK,
) -> Callable[[T], T]:
    def decorator(cls: T) -> T:
        task_name = name or camel_to_snake(cls.__qualname__)
        method = getattr(cls, method_name)
        setattr(
            cls,
            method_name,
            entity_method(name=task_name, version=version, tlp_span_kind=tlp_span_kind)(
                method
            ),
        )
        return cls

    return decorator



from collections.abc import Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from opentelemetry.util.types import Attributes

# Set of LLM provider names used in OpenTelemetry attributes
# See https://opentelemetry.io/docs/specs/semconv/registry/attributes/gen-ai/#gen-ai-provider-name
LLM_PROVIDER_NAMES = {
    "anthropic",
    "aws.bedrock",
    "azure.ai.inference",
    "azure.ai.openai",
    "cohere",
    "deepseek",
    "gcp.gemini",
    "gcp.gen_ai",
    "gcp.vertex_ai",
    "groq",
    "ibm.watsonx.ai",
    "mistral_ai",
    "openai",
    "perplexity",
    "x_ai",
}

CLOUD_PROVIDERS_NAMES = {"metadata.google.internal"}


class NoopExporter(SpanExporter):
    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        return SpanExportResult.SUCCESS


class FilteringExporter(SpanExporter):
    def __init__(
        self,
        inner: SpanExporter,
        llm_provider_blocklist: set[str] | None = None,
        attribute_filter: dict[str, set[str]] | None = None,
    ):
        self._inner = inner
        self._llm_provider_blocklist = llm_provider_blocklist or set()
        self._attribute_filter = attribute_filter or {}

    def _filter_based_on_llm_provider(self, attrs: Attributes) -> bool:
        attrs = attrs or {}
        provider = attrs.get("gen_ai.system", "") or attrs.get(
            "gen_ai.provider.name", ""
        )
        return provider not in self._llm_provider_blocklist

    def _filter_based_on_attrs(self, attrs: Attributes) -> bool:
        attrs = attrs or {}
        for name, values in self._attribute_filter.items():
            attr_value = attrs.get(name, "")
            for pattern in values:
                if pattern in str(attr_value):
                    return False
        return True

    def export(self, spans: Sequence[ReadableSpan]):
        keep: list[ReadableSpan] = []
        for s in spans:
            if self._filter_based_on_attrs(
                s.attributes
            ) and self._filter_based_on_llm_provider(s.attributes):
                keep.append(s)

        return SpanExportResult.SUCCESS if not keep else self._inner.export(keep)

    def shutdown(self):
        self._inner.shutdown()

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return self._inner.force_flush(timeout_millis)


class LogScopeExporter(SpanExporter):
    def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        for s in spans:
            scope = getattr(s, "instrumentation_scope", None)
            print(
                "SCOPE:",
                getattr(scope, "name", ""),
                "SPAN:",
                s.name,
                "ATTRS:",
                s.attributes,
            )
        return SpanExportResult.SUCCESS



from collections.abc import Sequence
from uuid import uuid4

from litellm.types.llms.openai import ChatCompletionAnnotation
from pydantic import BaseModel, Field

from ..typing.tool import ToolCall


class Reasoning(BaseModel):
    summaries: list[str]
    encrypted_content: str | None = None


class TextOutput(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid4())[:8])
    content: str
    annotations: Sequence[ChatCompletionAnnotation] | None = None
    refusal: str | None = None


CompletionItem = ToolCall | Reasoning | TextOutput



from ...errors import CompletionError
from ...typing.completion import Completion, Usage
from . import OpenAICompletion, OpenAIUsage
from .message_converters import from_api_assistant_message


def from_api_completion_usage(api_usage: OpenAIUsage) -> Usage:
    reasoning_tokens = None
    cached_tokens = None

    if api_usage.completion_tokens_details is not None:
        reasoning_tokens = api_usage.completion_tokens_details.reasoning_tokens
    if api_usage.prompt_tokens_details is not None:
        cached_tokens = api_usage.prompt_tokens_details.cached_tokens

    input_tokens = max(api_usage.prompt_tokens - (cached_tokens or 0), 0)
    output_tokens = max(api_usage.completion_tokens - (reasoning_tokens or 0), 0)

    return Usage(
        input_tokens=input_tokens,
        output_tokens=output_tokens,
        reasoning_tokens=reasoning_tokens,
        cached_tokens=cached_tokens,
    )


# response completed event and then only we can get response class


def from_api_completion(
    api_completion: OpenAICompletion, name: str | None = None
) -> Completion:
    # Some providers return None for the choices when there is an error
    # TODO: add custom error types
    if api_completion.choices is None:  # type: ignore
        raise CompletionError(
            f"Completion API error: {getattr(api_completion, 'error', None)}"
        )

    if not api_completion.choices:
        raise CompletionError("No choices in completion")

    if len(api_completion.choices) > 1:
        raise CompletionError("Multiple choices are not supported")

    api_choice = api_completion.choices[0]

    # Some providers return None for the message when finish_reason is other than "stop"
    finish_reason = api_choice.finish_reason
    if api_choice.message is None:  # type: ignore
        raise CompletionError(
            f"API returned None for message with finish_reason: {finish_reason}"
        )

    message = from_api_assistant_message(api_choice.message, name=name)
    usage = (
        from_api_completion_usage(api_completion.usage)
        if api_completion.usage
        else None
    )

    return Completion(
        id=api_completion.id,
        created=api_completion.created,
        system_fingerprint=api_completion.system_fingerprint,
        message=message,
        finish_reason=finish_reason,
        logprobs=api_choice.logprobs,
        usage=usage,
        model=api_completion.model,
        name=name,
    )


def to_api_completion(completion: Completion) -> OpenAICompletion:
    raise NotImplementedError



from openai.types.responses import (
    ResponseFunctionToolCall,
    ResponseOutputItemDoneEvent,
    ResponseOutputMessage,
    ResponseReasoningItem,
)

from ...typing.completion_item import CompletionItem, Reasoning, TextOutput
from ...typing.message import ToolCall
from .message_converters import from_api_responses_output_message


def from_api_completion_item(
    api_chunk: ResponseOutputItemDoneEvent, name: str | None = None
) -> CompletionItem:
    item = api_chunk.item
    if isinstance(item, ResponseReasoningItem):
        summaries = [reasoning.text for reasoning in item.summary]
        encrypted_content = item.encrypted_content
        return Reasoning(summaries=summaries, encrypted_content=encrypted_content)
    if isinstance(item, ResponseFunctionToolCall):
        tool_call = ToolCall(
            tool_arguments=item.arguments, tool_name=item.name, id=item.call_id
        )
        return tool_call
    if isinstance(item, ResponseOutputMessage):
        content_item, annotations, refusal = from_api_responses_output_message(item)
        return TextOutput(
            content=" ".join(content_item), annotations=annotations, refusal=refusal
        )
    raise TypeError(
        f"Cannot convert event of type '{type(api_chunk).__name__}' to CompletionItem"
    )



import logging
from collections.abc import AsyncIterator, Sequence
from typing import Any, ClassVar, Generic, TypeVar, cast, final

from pydantic import TypeAdapter
from pydantic import ValidationError as PydanticValidationError
from typing_extensions import Protocol

from grasp_agents.errors import (
    PacketRoutingError,
    ProcInputValidationError,
    ProcOutputValidationError,
)
from grasp_agents.tracing_decorators import workflow

from ..packet import Packet
from ..run_context import CtxT, RunContext
from ..typing.events import Event, ProcPacketOutEvent, ProcPayloadOutEvent
from ..typing.io import InT, OutT, ProcName
from ..utils.callbacks import is_method_overridden
from .base_processor import BaseProcessor, with_retry

logger = logging.getLogger(__name__)


_OutT_contra = TypeVar("_OutT_contra", contravariant=True)


class RecipientSelector(Protocol[_OutT_contra, CtxT]):
    def __call__(
        self, output: _OutT_contra, *, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]: ...


class Processor(BaseProcessor[InT, OutT, CtxT], Generic[InT, OutT, CtxT]):
    """
    Processor that can have different numbers of inputs and outputs, allowing for an
    arbitrary mapping between them.
    """

    _generic_arg_to_instance_attr_map: ClassVar[dict[int, str]] = {
        0: "_in_type",
        1: "_out_type",
    }

    def validate_inputs(
        self,
        call_id: str,
        chat_inputs: Any | None = None,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
    ) -> list[InT] | None:
        err_kwargs = {"proc_name": self.name, "call_id": call_id}

        num_non_null_inputs = sum(
            x is not None for x in [chat_inputs, in_args, in_packet]
        )

        if num_non_null_inputs > 1:
            raise ProcInputValidationError(
                message="Only one of chat_inputs, in_args, or in_message must be provided",
                **err_kwargs,
            )
        if self.in_type is not type(None) and num_non_null_inputs == 0:
            raise ProcInputValidationError(
                message="One of chat_inputs, in_args, or in_message must be provided",
                **err_kwargs,
            )

        if in_packet is not None and not in_packet.payloads:
            raise ProcInputValidationError(
                message="in_packet must contain at least one payload", **err_kwargs
            )
        if in_args is not None and not in_args:
            raise ProcInputValidationError(
                message="in_args must contain at least one argument", **err_kwargs
            )

        if chat_inputs is not None:
            # 1) chat_inputs are provided -> no need to validate further
            return None

        resolved_args: list[InT]

        if isinstance(in_args, self.in_type):
            # 2) Single in_args of correct type is provided
            resolved_args = [in_args]

        elif isinstance(in_args, list):
            # 3) List of in_args is provided
            resolved_args = cast("list[InT]", in_args)

        elif in_args is not None:
            raise ProcInputValidationError(
                message=f"in_args are neither of type {self.in_type} "
                f"nor a list of {self.in_type}.",
                **err_kwargs,
            )

        else:
            # 4) in_packet is provided
            resolved_args = list(cast("Packet[InT]", in_packet).payloads)

        try:
            for args in resolved_args:
                TypeAdapter(self._in_type).validate_python(args)
        except PydanticValidationError as err:
            raise ProcInputValidationError(message=str(err), **err_kwargs) from err

        return resolved_args

    def validate_output(self, out_payload: OutT, call_id: str) -> OutT:
        if out_payload is None:
            return out_payload

        try:
            return TypeAdapter(self.out_type).validate_python(out_payload)
        except PydanticValidationError as err:
            raise ProcOutputValidationError(
                schema=self.out_type, proc_name=self.name, call_id=call_id
            ) from err

    def _validate_recipients(
        self, recipients: Sequence[ProcName] | None, call_id: str
    ) -> None:
        for r in recipients or []:
            if r not in (self.recipients or []):
                raise PacketRoutingError(
                    proc_name=self.name,
                    call_id=call_id,
                    selected_recipient=r,
                    allowed_recipients=cast("list[str]", self.recipients),
                )

    def select_recipients_impl(
        self, output: OutT, *, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]:
        raise NotImplementedError

    def add_recipient_selector(
        self, func: RecipientSelector[OutT, CtxT]
    ) -> RecipientSelector[OutT, CtxT]:
        self.select_recipients_impl = func

        return func

    @final
    def select_recipients(
        self, output: OutT, ctx: RunContext[CtxT], call_id: str
    ) -> Sequence[ProcName]:
        base_cls = BaseProcessor[Any, Any, Any]
        if is_method_overridden("select_recipients_impl", self, base_cls):
            recipients = self.select_recipients_impl(
                output=output, ctx=ctx, call_id=call_id
            )
            self._validate_recipients(recipients, call_id=call_id)
            return recipients

        # self.processor is not None because otherwise we do not call this method
        return cast("list[ProcName]", self.recipients)

    async def _process(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> list[OutT]:
        """
        Process a list of inputs and return a list of outputs. The length of the
        output list can be different from the input list.
        """
        return cast("list[OutT]", in_args)

    async def _process_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_args: list[InT] | None = None,
        call_id: str,
        ctx: RunContext[CtxT],
    ) -> AsyncIterator[Event[Any]]:
        outputs = await self._process(
            chat_inputs=chat_inputs, in_args=in_args, call_id=call_id, ctx=ctx
        )
        for output in outputs:
            yield ProcPayloadOutEvent(data=output, src_name=self.name, call_id=call_id)

    def _build_packet(
        self, outputs: list[OutT], call_id: str, ctx: RunContext[CtxT]
    ) -> Packet[OutT]:
        for output in outputs:
            self.validate_output(output, call_id=call_id)

        routings: list[Sequence[ProcName]] | None = []
        if self.recipients is not None:
            for output in outputs:
                routings.append(
                    self.select_recipients(output=output, ctx=ctx, call_id=call_id)
                )

        joined_routing = [r for r in routings] if routings else None

        return Packet(sender=self.name, payloads=outputs, routing=joined_routing)

    @final
    @workflow(name="processor")  # type: ignore
    @with_retry
    async def run_stream(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> AsyncIterator[Event[Any]]:
        ctx = ctx or RunContext[CtxT](state=None)  # type: ignore
        call_id = self.generate_call_id(call_id)

        val_in_args = self.validate_inputs(
            call_id=call_id,
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
        )

        outputs: list[OutT] = []
        async for event in self._process_stream(
            chat_inputs=chat_inputs,
            in_args=val_in_args,
            call_id=call_id,
            ctx=ctx,
        ):
            if isinstance(event, ProcPayloadOutEvent) and event.src_name == self.name:
                # Collect payloads of the processor's own output events
                # This makes sure subprocessor events are yielded but their
                # payloads are not included in the final output packet
                outputs.append(event.data)
            else:
                yield event

        # 1) Combine payloads
        # 2) Obtain recipients per payload
        # 3) Form a packet
        # 4) Yield a ProcPacketOutEvent
        out_packet = self._build_packet(outputs=outputs, call_id=call_id, ctx=ctx)
        yield ProcPacketOutEvent(
            id=out_packet.id, data=out_packet, src_name=self.name, call_id=call_id
        )

    async def run(
        self,
        chat_inputs: Any | None = None,
        *,
        in_packet: Packet[InT] | None = None,
        in_args: InT | list[InT] | None = None,
        call_id: str | None = None,
        ctx: RunContext[CtxT] | None = None,
    ) -> Packet[OutT]:
        result = None

        async for event in self.run_stream(
            chat_inputs=chat_inputs,
            in_packet=in_packet,
            in_args=in_args,
            call_id=call_id,
            ctx=ctx,
        ):
            if result is not None:
                # Need to drain the event stream for OTel to work properly
                continue

            if isinstance(event, ProcPacketOutEvent) and event.src_name == self.name:
                result = event.data

        if result is None:
            raise RuntimeError("Processor run did not yield a ProcPacketOutputEvent")

        return result
